from datetime import datetime
import json
from typing import Any, Dict, IO, List, Optional, Union

import click
from click import ClickException
import tabulate
import yaml
from yaml.loader import SafeLoader

from anyscale.anyscale_pydantic import (
    BaseModel,
    Field,
    root_validator,
    validator,
)
from anyscale.cli_logger import BlockLogger
from anyscale.cloud import get_cloud_id_and_name
from anyscale.cluster_compute import get_cluster_compute_from_name
from anyscale.conf import IDLE_TIMEOUT_DEFAULT_MINUTES
from anyscale.controllers.base_controller import BaseController
from anyscale.sdk.anyscale_client.models.cluster_compute_config import (
    ClusterComputeConfig,
)
from anyscale.sdk.anyscale_client.models.cluster_computes_query import (
    ClusterComputesQuery,
)
from anyscale.sdk.anyscale_client.models.clustercompute_list_response import (
    ClustercomputeListResponse,
)
from anyscale.sdk.anyscale_client.models.create_cluster_compute import (
    CreateClusterCompute,
)
from anyscale.sdk.anyscale_client.models.create_cluster_compute_config import (
    CreateClusterComputeConfig,
)
from anyscale.sdk.anyscale_client.models.text_query import TextQuery
from anyscale.util import get_endpoint
from anyscale.utils.entity_arg_utils import EntityType, IdBasedEntity, NameBasedEntity


log = BlockLogger()  # Anyscale CLI Logger


class CreateClusterComputeConfigModel(BaseModel):
    """
    Schema for ClusterClusterCompute that is accepted from the CLI. Supports specifying
    `cloud` name instead of `cloud_id`, and preprocesses so fields have the correct
    values to call POST `/ext/v0/cluster_computes`.
    """

    cloud_id: Optional[str] = Field(
        None, description="The ID of the Anyscale cloud to use for launching Clusters.",
    )
    cloud: Optional[str] = Field(
        None,
        description="The name of the Anyscale cloud to use for launching Clusters.",
    )
    max_workers: Optional[int] = Field(
        None, description="Desired limit on total running workers for this Cluster."
    )

    allowed_azs: Optional[List[str]] = Field(
        None,
        description='The availability zones that clusters are allowed to be launched in, e.g. "us-west-2a". If not specified, any AZ may be used.',
    )

    head_node_type: Any = Field(
        ..., description="Node configuration to use for the head node. ",
    )

    worker_node_types: List[Any] = Field(
        ..., description="A list of node types to use for worker nodes. "
    )

    aws: Optional[Any] = Field(
        None,
        description="Fields specific to AWS node types.",
        alias="aws_advanced_configurations_json",
    )

    gcp: Optional[Any] = Field(
        None,
        description="Fields specific to GCP node types.",
        alias="gcp_advanced_configurations_json",
    )

    azure: Optional[Any] = Field(
        None, description="Fields specific to Azure node types.",
    )

    maximum_uptime_minutes: Optional[int] = Field(
        None,
        description="If set to a positive number, Anyscale will terminate the cluster this many minutes after cluster start.",
    )

    idle_termination_minutes: int = Field(
        IDLE_TIMEOUT_DEFAULT_MINUTES,
        description=(
            "If set to a positive number, Anyscale will terminate the cluster this many minutes after the cluster is idle. "
            "Idle time is defined as the time during which a Cluster is not running a user command. "
            "Time spent running commands on Jupyter or ssh is still considered 'idle'. "
            "To disable, set this field to 0."
        ),
        ge=0,
    )

    @validator("allowed_azs")
    def check_any_option(cls, v: Optional[List[str]]):
        if not v:
            return v
        if "any" in v and len(v) > 1:
            raise ValueError("Cannot specify 'any' and other AZs.")
        return v

    @root_validator
    def fill_cloud_id(cls: Any, values: Any) -> Any:
        cloud_id, cloud = (
            values.get("cloud_id"),
            values.get("cloud"),
        )
        if cloud_id and cloud:
            raise click.ClickException(
                "Only one of `cloud_id` or `cloud` can be provided in the cluster compute config file. "
            )
        if cloud:
            cloud_id, _ = get_cloud_id_and_name(
                api_client=None, cloud_id=None, cloud_name=cloud
            )
            values["cloud_id"] = cloud_id
        elif not cloud_id:
            raise click.ClickException(
                "Please provide either `cloud_id` or `cloud` in the cluster compute config file. "
            )
        return values

    class Config:
        # This setting will cause pydantic (and fastAPI) to accept the field name as well as the alias
        # (eg. "cpu" or "CPU"). But note that openapi does not respect this setting.
        # Generated models will require the alias.
        allow_population_by_field_name = True


class ComputeConfigController(BaseController):
    """
    This controller powers functionalities related to Anyscale
    cluster compute configuration.
    """

    def __init__(
        self, log: Optional[BlockLogger] = None, initialize_auth_api_client: bool = True
    ):
        if log is None:
            log = BlockLogger()

        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.log.open_block("Output")

    def create(self, cluster_compute_file: IO[bytes], name: Optional[str]) -> None:
        """Builds a new cluster compute template
        If name is not provided, a default cluster-compute-name will be used and returned in the command output

        Information in output: Link to cluster compute in UI, cluster compute id
        """

        try:
            cluster_compute: Dict[str, Any] = yaml.load(
                cluster_compute_file, Loader=SafeLoader
            )
        except Exception as e:  # noqa: BLE001
            raise ClickException(f"Could not load cluster compute file: {e}")

        cluster_compute_config_model = CreateClusterComputeConfigModel(
            **cluster_compute
        )
        if cluster_compute_config_model.max_workers is not None:
            log.warning(
                "Warning: global `max_workers` is deprecated for Anyscale Ray 2.7+. Please use global resource max instead: https://docs.anyscale.com/configure/compute-configs/global-resource-min-max"
            )
        cluster_compute_config = CreateClusterComputeConfig(
            cloud_id=cluster_compute_config_model.cloud_id,
            max_workers=cluster_compute_config_model.max_workers,
            allowed_azs=cluster_compute_config_model.allowed_azs,
            region=None,
            head_node_type=cluster_compute_config_model.head_node_type,
            worker_node_types=cluster_compute_config_model.worker_node_types,
            aws_advanced_configurations_json=cluster_compute_config_model.aws,
            gcp_advanced_configurations_json=cluster_compute_config_model.gcp,
            maximum_uptime_minutes=cluster_compute_config_model.maximum_uptime_minutes,
            idle_termination_minutes=cluster_compute_config_model.idle_termination_minutes,
        )
        if name is None:
            name = f"cli-config-{datetime.now().isoformat()}"
        cluster_compute_response = self.anyscale_api_client.create_cluster_compute(
            CreateClusterCompute(name=name, config=cluster_compute_config)
        )
        created_cluster_compute = cluster_compute_response.result
        cluster_compute_id = created_cluster_compute.id
        cluster_compute_name = created_cluster_compute.name
        url = get_endpoint(f"/configurations/cluster-computes/{cluster_compute_id}")
        log.info(f"View this cluster compute at: {url}")
        log.info(f"Cluster compute id: {cluster_compute_id}")
        log.info(f"Cluster compute name: {cluster_compute_name}")

    def delete(
        self, cluster_compute_name: Optional[str], id: Optional[str]  # noqa: A002
    ) -> None:
        """
        TODO: shawnp. Remove once archive functionality is enabled on prod.
        Deletes the cluster compute with the given name or id.
        Exactly one of cluster_compute_name or id must be provided.
        """

        if int(bool(cluster_compute_name)) + int(bool(id)) != 1:
            raise ClickException(
                "Not deleted. Please provide exactly one of: cluster compute name, id."
            )

        if id:
            self.anyscale_api_client.get_cluster_compute(id)
            cluster_compute_id = id
        else:
            # find the cluster compute id from the name
            query = ClusterComputesQuery(name=TextQuery(equals=cluster_compute_name))
            query_response: ClustercomputeListResponse = self.anyscale_api_client.search_cluster_computes(
                query
            )
            cluster_results = query_response.results
            if len(cluster_results) != 1:
                raise ClickException(
                    f"Not deleted. No cluster compute template exists with the name {cluster_compute_name}."
                )
            cluster_compute_id = cluster_results[0].id
        self.anyscale_api_client.delete_cluster_compute(cluster_compute_id)
        log.info("Cluster compute deleted.")

    def archive(
        self, compute_config_entity: Union[IdBasedEntity, NameBasedEntity]
    ) -> None:
        """
        Archives the cluster compute with the given name or id.
        Exactly one of cluster_compute_name or id must be provided.
        """
        if compute_config_entity.type is EntityType.ID:
            compute_config = self.anyscale_api_client.get_compute_template(
                compute_config_entity.id
            ).result
            compute_config_id = compute_config_entity.id
            compute_config_name = compute_config.name
        else:
            compute_config = get_cluster_compute_from_name(
                compute_config_entity.name,
                self.api_client,
                compute_config_entity.version,
            )
            compute_config_id = compute_config.id
            compute_config_name = compute_config.name
        self.api_client.archive_compute_template_api_v2_compute_templates_compute_template_id_archive_post(
            compute_config_id
        )
        log.info(f"Successfully archived compute config: {compute_config_name}.")

    def list(
        self,
        cluster_compute_name: Optional[str],
        cluster_compute_id: Optional[str],
        include_shared: bool,
        max_items: int,
    ) -> None:
        cluster_compute_list = []
        if cluster_compute_id:
            cluster_compute_list = [
                self.anyscale_api_client.get_cluster_compute(cluster_compute_id).result
            ]
        elif cluster_compute_name:
            cluster_compute_list = self.anyscale_api_client.search_cluster_computes(
                {"name": {"equals": cluster_compute_name}, "paging": {"count": 1}}
            ).results
        else:
            creator_id = (
                self.api_client.get_user_info_api_v2_userinfo_get().result.id
                if not include_shared
                else None
            )
            resp = self.anyscale_api_client.search_cluster_computes(
                {"creator_id": creator_id, "paging": {"count": 20}}
            )
            cluster_compute_list.extend(resp.results)
            paging_token = resp.metadata.next_paging_token
            has_more = (paging_token is not None) and (
                len(cluster_compute_list) < max_items
            )
            while has_more:
                resp = self.anyscale_api_client.search_cluster_computes(
                    {
                        "creator_id": creator_id,
                        "paging": {"count": 20, "paging_token": paging_token},
                    }
                )
                cluster_compute_list.extend(resp.results)
                paging_token = resp.metadata.next_paging_token
                has_more = (paging_token is not None) and (
                    len(cluster_compute_list) < max_items
                )
            cluster_compute_list = cluster_compute_list[:max_items]

        cluster_compute_table = [
            [
                cluster_compute.id,
                cluster_compute.name,
                self.anyscale_api_client.get_cloud(
                    cluster_compute.config.cloud_id
                ).result.name
                if cluster_compute.config.cloud_id
                else None,
                cluster_compute.last_modified_at.strftime("%m/%d/%Y, %H:%M:%S"),
                get_endpoint(f"configurations/cluster-computes/{cluster_compute.id}"),
            ]
            for cluster_compute in cluster_compute_list
        ]

        table = tabulate.tabulate(
            cluster_compute_table,
            headers=["ID", "NAME", "CLOUD", "LAST MODIFIED AT", "URL"],
            tablefmt="plain",
        )
        print(f"Cluster computes:\n{table}")

    def get(
        self,
        cluster_compute_name: Optional[str],
        cluster_compute_id: Optional[str],
        cluster_compute_version: Optional[int] = None,
    ) -> None:
        if (
            int(cluster_compute_name is not None) + int(cluster_compute_id is not None)
            != 1
        ):
            raise click.ClickException(
                "Please only provide one of `cluster-compute-name` or `--id`."
            )
        if cluster_compute_name:
            cluster_compute_id = get_cluster_compute_from_name(
                cluster_compute_name, self.api_client, cluster_compute_version
            ).id
        compute_config: ClusterComputeConfig = self.anyscale_api_client.get_cluster_compute(
            cluster_compute_id
        ).result.config

        compute_config_dict = {
            key: val for key, val in compute_config.to_dict().items() if val is not None
        }
        formatted_config_json = json.dumps(compute_config_dict, indent=2)
        print(formatted_config_json)



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedSchedule(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'description': 'str',
        'project_id': 'str',
        'config': 'ProductionJobConfig',
        'schedule': 'ScheduleConfig',
        'id': 'str',
        'created_at': 'datetime',
        'updated_at': 'datetime',
        'creator_id': 'str',
        'next_trigger_at': 'datetime',
        'controller_user_id': 'str',
        'project': 'MiniProject',
        'creator': 'MiniUser',
        'last_executions': 'list[MiniProductionJob]'
    }

    attribute_map = {
        'name': 'name',
        'description': 'description',
        'project_id': 'project_id',
        'config': 'config',
        'schedule': 'schedule',
        'id': 'id',
        'created_at': 'created_at',
        'updated_at': 'updated_at',
        'creator_id': 'creator_id',
        'next_trigger_at': 'next_trigger_at',
        'controller_user_id': 'controller_user_id',
        'project': 'project',
        'creator': 'creator',
        'last_executions': 'last_executions'
    }

    def __init__(self, name=None, description=None, project_id=None, config=None, schedule=None, id=None, created_at=None, updated_at=None, creator_id=None, next_trigger_at=None, controller_user_id=None, project=None, creator=None, last_executions=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedSchedule - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._description = None
        self._project_id = None
        self._config = None
        self._schedule = None
        self._id = None
        self._created_at = None
        self._updated_at = None
        self._creator_id = None
        self._next_trigger_at = None
        self._controller_user_id = None
        self._project = None
        self._creator = None
        self._last_executions = None
        self.discriminator = None

        self.name = name
        if description is not None:
            self.description = description
        self.project_id = project_id
        self.config = config
        self.schedule = schedule
        self.id = id
        self.created_at = created_at
        self.updated_at = updated_at
        self.creator_id = creator_id
        if next_trigger_at is not None:
            self.next_trigger_at = next_trigger_at
        if controller_user_id is not None:
            self.controller_user_id = controller_user_id
        self.project = project
        self.creator = creator
        self.last_executions = last_executions

    @property
    def name(self):
        """Gets the name of this DecoratedSchedule.  # noqa: E501

        Name of the job  # noqa: E501

        :return: The name of this DecoratedSchedule.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this DecoratedSchedule.

        Name of the job  # noqa: E501

        :param name: The name of this DecoratedSchedule.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def description(self):
        """Gets the description of this DecoratedSchedule.  # noqa: E501

        Description of the job  # noqa: E501

        :return: The description of this DecoratedSchedule.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this DecoratedSchedule.

        Description of the job  # noqa: E501

        :param description: The description of this DecoratedSchedule.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def project_id(self):
        """Gets the project_id of this DecoratedSchedule.  # noqa: E501

        Id of the project this job will start clusters in  # noqa: E501

        :return: The project_id of this DecoratedSchedule.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this DecoratedSchedule.

        Id of the project this job will start clusters in  # noqa: E501

        :param project_id: The project_id of this DecoratedSchedule.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def config(self):
        """Gets the config of this DecoratedSchedule.  # noqa: E501

        The config that was used to create this job  # noqa: E501

        :return: The config of this DecoratedSchedule.  # noqa: E501
        :rtype: ProductionJobConfig
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this DecoratedSchedule.

        The config that was used to create this job  # noqa: E501

        :param config: The config of this DecoratedSchedule.  # noqa: E501
        :type: ProductionJobConfig
        """
        if self.local_vars_configuration.client_side_validation and config is None:  # noqa: E501
            raise ValueError("Invalid value for `config`, must not be `None`")  # noqa: E501

        self._config = config

    @property
    def schedule(self):
        """Gets the schedule of this DecoratedSchedule.  # noqa: E501

        The configuration for this schedule  # noqa: E501

        :return: The schedule of this DecoratedSchedule.  # noqa: E501
        :rtype: ScheduleConfig
        """
        return self._schedule

    @schedule.setter
    def schedule(self, schedule):
        """Sets the schedule of this DecoratedSchedule.

        The configuration for this schedule  # noqa: E501

        :param schedule: The schedule of this DecoratedSchedule.  # noqa: E501
        :type: ScheduleConfig
        """
        if self.local_vars_configuration.client_side_validation and schedule is None:  # noqa: E501
            raise ValueError("Invalid value for `schedule`, must not be `None`")  # noqa: E501

        self._schedule = schedule

    @property
    def id(self):
        """Gets the id of this DecoratedSchedule.  # noqa: E501

        The id of this job  # noqa: E501

        :return: The id of this DecoratedSchedule.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedSchedule.

        The id of this job  # noqa: E501

        :param id: The id of this DecoratedSchedule.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedSchedule.  # noqa: E501

        The time this job was created  # noqa: E501

        :return: The created_at of this DecoratedSchedule.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedSchedule.

        The time this job was created  # noqa: E501

        :param created_at: The created_at of this DecoratedSchedule.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def updated_at(self):
        """Gets the updated_at of this DecoratedSchedule.  # noqa: E501

        The time this job was last updated  # noqa: E501

        :return: The updated_at of this DecoratedSchedule.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_at

    @updated_at.setter
    def updated_at(self, updated_at):
        """Sets the updated_at of this DecoratedSchedule.

        The time this job was last updated  # noqa: E501

        :param updated_at: The updated_at of this DecoratedSchedule.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_at is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_at`, must not be `None`")  # noqa: E501

        self._updated_at = updated_at

    @property
    def creator_id(self):
        """Gets the creator_id of this DecoratedSchedule.  # noqa: E501

        The id of the user who created this job  # noqa: E501

        :return: The creator_id of this DecoratedSchedule.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this DecoratedSchedule.

        The id of the user who created this job  # noqa: E501

        :param creator_id: The creator_id of this DecoratedSchedule.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def next_trigger_at(self):
        """Gets the next_trigger_at of this DecoratedSchedule.  # noqa: E501

        The next UTC timestamp at which this cron job will trigger.  # noqa: E501

        :return: The next_trigger_at of this DecoratedSchedule.  # noqa: E501
        :rtype: datetime
        """
        return self._next_trigger_at

    @next_trigger_at.setter
    def next_trigger_at(self, next_trigger_at):
        """Sets the next_trigger_at of this DecoratedSchedule.

        The next UTC timestamp at which this cron job will trigger.  # noqa: E501

        :param next_trigger_at: The next_trigger_at of this DecoratedSchedule.  # noqa: E501
        :type: datetime
        """

        self._next_trigger_at = next_trigger_at

    @property
    def controller_user_id(self):
        """Gets the controller_user_id of this DecoratedSchedule.  # noqa: E501

        The id of the user that started, paused or resumed the schedule most recently.  # noqa: E501

        :return: The controller_user_id of this DecoratedSchedule.  # noqa: E501
        :rtype: str
        """
        return self._controller_user_id

    @controller_user_id.setter
    def controller_user_id(self, controller_user_id):
        """Sets the controller_user_id of this DecoratedSchedule.

        The id of the user that started, paused or resumed the schedule most recently.  # noqa: E501

        :param controller_user_id: The controller_user_id of this DecoratedSchedule.  # noqa: E501
        :type: str
        """

        self._controller_user_id = controller_user_id

    @property
    def project(self):
        """Gets the project of this DecoratedSchedule.  # noqa: E501

        The project in which this schedule lives  # noqa: E501

        :return: The project of this DecoratedSchedule.  # noqa: E501
        :rtype: MiniProject
        """
        return self._project

    @project.setter
    def project(self, project):
        """Sets the project of this DecoratedSchedule.

        The project in which this schedule lives  # noqa: E501

        :param project: The project of this DecoratedSchedule.  # noqa: E501
        :type: MiniProject
        """
        if self.local_vars_configuration.client_side_validation and project is None:  # noqa: E501
            raise ValueError("Invalid value for `project`, must not be `None`")  # noqa: E501

        self._project = project

    @property
    def creator(self):
        """Gets the creator of this DecoratedSchedule.  # noqa: E501

        The creator of this schedule  # noqa: E501

        :return: The creator of this DecoratedSchedule.  # noqa: E501
        :rtype: MiniUser
        """
        return self._creator

    @creator.setter
    def creator(self, creator):
        """Sets the creator of this DecoratedSchedule.

        The creator of this schedule  # noqa: E501

        :param creator: The creator of this DecoratedSchedule.  # noqa: E501
        :type: MiniUser
        """
        if self.local_vars_configuration.client_side_validation and creator is None:  # noqa: E501
            raise ValueError("Invalid value for `creator`, must not be `None`")  # noqa: E501

        self._creator = creator

    @property
    def last_executions(self):
        """Gets the last_executions of this DecoratedSchedule.  # noqa: E501


        :return: The last_executions of this DecoratedSchedule.  # noqa: E501
        :rtype: list[MiniProductionJob]
        """
        return self._last_executions

    @last_executions.setter
    def last_executions(self, last_executions):
        """Sets the last_executions of this DecoratedSchedule.


        :param last_executions: The last_executions of this DecoratedSchedule.  # noqa: E501
        :type: list[MiniProductionJob]
        """
        if self.local_vars_configuration.client_side_validation and last_executions is None:  # noqa: E501
            raise ValueError("Invalid value for `last_executions`, must not be `None`")  # noqa: E501

        self._last_executions = last_executions

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedSchedule):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedSchedule):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedServiceEventAPIModel(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'created_at': 'datetime',
        'event_type': 'ServiceEventType',
        'current_state': 'ServiceEventCurrentState',
        'goal_state': 'ServiceGoalStates',
        'level': 'ServiceEventLevel',
        'message': 'str',
        'origin': 'ServiceEventOrigin',
        'service_id': 'str',
        'service_version_name': 'str',
        'service_version_id': 'str',
        'cluster_id': 'str',
        'goal_state_initiator': 'MiniUser'
    }

    attribute_map = {
        'id': 'id',
        'created_at': 'created_at',
        'event_type': 'event_type',
        'current_state': 'current_state',
        'goal_state': 'goal_state',
        'level': 'level',
        'message': 'message',
        'origin': 'origin',
        'service_id': 'service_id',
        'service_version_name': 'service_version_name',
        'service_version_id': 'service_version_id',
        'cluster_id': 'cluster_id',
        'goal_state_initiator': 'goal_state_initiator'
    }

    def __init__(self, id=None, created_at=None, event_type=None, current_state=None, goal_state=None, level=None, message=None, origin=None, service_id=None, service_version_name=None, service_version_id=None, cluster_id=None, goal_state_initiator=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedServiceEventAPIModel - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._created_at = None
        self._event_type = None
        self._current_state = None
        self._goal_state = None
        self._level = None
        self._message = None
        self._origin = None
        self._service_id = None
        self._service_version_name = None
        self._service_version_id = None
        self._cluster_id = None
        self._goal_state_initiator = None
        self.discriminator = None

        self.id = id
        self.created_at = created_at
        self.event_type = event_type
        if current_state is not None:
            self.current_state = current_state
        if goal_state is not None:
            self.goal_state = goal_state
        if level is not None:
            self.level = level
        if message is not None:
            self.message = message
        self.origin = origin
        if service_id is not None:
            self.service_id = service_id
        if service_version_name is not None:
            self.service_version_name = service_version_name
        if service_version_id is not None:
            self.service_version_id = service_version_id
        if cluster_id is not None:
            self.cluster_id = cluster_id
        if goal_state_initiator is not None:
            self.goal_state_initiator = goal_state_initiator

    @property
    def id(self):
        """Gets the id of this DecoratedServiceEventAPIModel.  # noqa: E501

        Id of the Service event  # noqa: E501

        :return: The id of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedServiceEventAPIModel.

        Id of the Service event  # noqa: E501

        :param id: The id of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedServiceEventAPIModel.  # noqa: E501

        The last time the state of this job was updated. This includes updates to the state and to the goal state  # noqa: E501

        :return: The created_at of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedServiceEventAPIModel.

        The last time the state of this job was updated. This includes updates to the state and to the goal state  # noqa: E501

        :param created_at: The created_at of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def event_type(self):
        """Gets the event_type of this DecoratedServiceEventAPIModel.  # noqa: E501

        Type of Service Event  # noqa: E501

        :return: The event_type of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: ServiceEventType
        """
        return self._event_type

    @event_type.setter
    def event_type(self, event_type):
        """Sets the event_type of this DecoratedServiceEventAPIModel.

        Type of Service Event  # noqa: E501

        :param event_type: The event_type of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: ServiceEventType
        """
        if self.local_vars_configuration.client_side_validation and event_type is None:  # noqa: E501
            raise ValueError("Invalid value for `event_type`, must not be `None`")  # noqa: E501

        self._event_type = event_type

    @property
    def current_state(self):
        """Gets the current_state of this DecoratedServiceEventAPIModel.  # noqa: E501

        Present only for Service events.  # noqa: E501

        :return: The current_state of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: ServiceEventCurrentState
        """
        return self._current_state

    @current_state.setter
    def current_state(self, current_state):
        """Sets the current_state of this DecoratedServiceEventAPIModel.

        Present only for Service events.  # noqa: E501

        :param current_state: The current_state of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: ServiceEventCurrentState
        """

        self._current_state = current_state

    @property
    def goal_state(self):
        """Gets the goal_state of this DecoratedServiceEventAPIModel.  # noqa: E501

        Present only for Service events.  # noqa: E501

        :return: The goal_state of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: ServiceGoalStates
        """
        return self._goal_state

    @goal_state.setter
    def goal_state(self, goal_state):
        """Sets the goal_state of this DecoratedServiceEventAPIModel.

        Present only for Service events.  # noqa: E501

        :param goal_state: The goal_state of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: ServiceGoalStates
        """

        self._goal_state = goal_state

    @property
    def level(self):
        """Gets the level of this DecoratedServiceEventAPIModel.  # noqa: E501


        :return: The level of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: ServiceEventLevel
        """
        return self._level

    @level.setter
    def level(self, level):
        """Sets the level of this DecoratedServiceEventAPIModel.


        :param level: The level of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: ServiceEventLevel
        """

        self._level = level

    @property
    def message(self):
        """Gets the message of this DecoratedServiceEventAPIModel.  # noqa: E501


        :return: The message of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this DecoratedServiceEventAPIModel.


        :param message: The message of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: str
        """

        self._message = message

    @property
    def origin(self):
        """Gets the origin of this DecoratedServiceEventAPIModel.  # noqa: E501

        The origin of the Service events.  # noqa: E501

        :return: The origin of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: ServiceEventOrigin
        """
        return self._origin

    @origin.setter
    def origin(self, origin):
        """Sets the origin of this DecoratedServiceEventAPIModel.

        The origin of the Service events.  # noqa: E501

        :param origin: The origin of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: ServiceEventOrigin
        """
        if self.local_vars_configuration.client_side_validation and origin is None:  # noqa: E501
            raise ValueError("Invalid value for `origin`, must not be `None`")  # noqa: E501

        self._origin = origin

    @property
    def service_id(self):
        """Gets the service_id of this DecoratedServiceEventAPIModel.  # noqa: E501

        The id of the Service v2. Present only for Service events.  # noqa: E501

        :return: The service_id of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._service_id

    @service_id.setter
    def service_id(self, service_id):
        """Sets the service_id of this DecoratedServiceEventAPIModel.

        The id of the Service v2. Present only for Service events.  # noqa: E501

        :param service_id: The service_id of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: str
        """

        self._service_id = service_id

    @property
    def service_version_name(self):
        """Gets the service_version_name of this DecoratedServiceEventAPIModel.  # noqa: E501

        The name of the associated Service Version. Present only for Service Version event  # noqa: E501

        :return: The service_version_name of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._service_version_name

    @service_version_name.setter
    def service_version_name(self, service_version_name):
        """Sets the service_version_name of this DecoratedServiceEventAPIModel.

        The name of the associated Service Version. Present only for Service Version event  # noqa: E501

        :param service_version_name: The service_version_name of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: str
        """

        self._service_version_name = service_version_name

    @property
    def service_version_id(self):
        """Gets the service_version_id of this DecoratedServiceEventAPIModel.  # noqa: E501

        The id of the associated Service Version. Present only for Service Version event  # noqa: E501

        :return: The service_version_id of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._service_version_id

    @service_version_id.setter
    def service_version_id(self, service_version_id):
        """Sets the service_version_id of this DecoratedServiceEventAPIModel.

        The id of the associated Service Version. Present only for Service Version event  # noqa: E501

        :param service_version_id: The service_version_id of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: str
        """

        self._service_version_id = service_version_id

    @property
    def cluster_id(self):
        """Gets the cluster_id of this DecoratedServiceEventAPIModel.  # noqa: E501

        The associated cluster id.Present only for Service Version event and only when the cluster is associated with the event.  # noqa: E501

        :return: The cluster_id of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: str
        """
        return self._cluster_id

    @cluster_id.setter
    def cluster_id(self, cluster_id):
        """Sets the cluster_id of this DecoratedServiceEventAPIModel.

        The associated cluster id.Present only for Service Version event and only when the cluster is associated with the event.  # noqa: E501

        :param cluster_id: The cluster_id of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: str
        """

        self._cluster_id = cluster_id

    @property
    def goal_state_initiator(self):
        """Gets the goal_state_initiator of this DecoratedServiceEventAPIModel.  # noqa: E501

        The user who initiated the service goal state change. Present only for goal state changes.  # noqa: E501

        :return: The goal_state_initiator of this DecoratedServiceEventAPIModel.  # noqa: E501
        :rtype: MiniUser
        """
        return self._goal_state_initiator

    @goal_state_initiator.setter
    def goal_state_initiator(self, goal_state_initiator):
        """Sets the goal_state_initiator of this DecoratedServiceEventAPIModel.

        The user who initiated the service goal state change. Present only for goal state changes.  # noqa: E501

        :param goal_state_initiator: The goal_state_initiator of this DecoratedServiceEventAPIModel.  # noqa: E501
        :type: MiniUser
        """

        self._goal_state_initiator = goal_state_initiator

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedServiceEventAPIModel):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedServiceEventAPIModel):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class UpdateCloud(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'config': 'CloudConfig',
        'auto_add_user': 'bool'
    }

    attribute_map = {
        'config': 'config',
        'auto_add_user': 'auto_add_user'
    }

    def __init__(self, config=None, auto_add_user=None, local_vars_configuration=None):  # noqa: E501
        """UpdateCloud - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._config = None
        self._auto_add_user = None
        self.discriminator = None

        if config is not None:
            self.config = config
        if auto_add_user is not None:
            self.auto_add_user = auto_add_user

    @property
    def config(self):
        """Gets the config of this UpdateCloud.  # noqa: E501

        Additional configurable properties of this Cloud.  # noqa: E501

        :return: The config of this UpdateCloud.  # noqa: E501
        :rtype: CloudConfig
        """
        return self._config

    @config.setter
    def config(self, config):
        """Sets the config of this UpdateCloud.

        Additional configurable properties of this Cloud.  # noqa: E501

        :param config: The config of this UpdateCloud.  # noqa: E501
        :type: CloudConfig
        """

        self._config = config

    @property
    def auto_add_user(self):
        """Gets the auto_add_user of this UpdateCloud.  # noqa: E501

        Whether all users in the organization should be automatically added to this cloud. This field is only relevant for organizations with cloud isolation enabled, because all users in the organization automatically have access to all clouds if cloud isolation is not enabled.  # noqa: E501

        :return: The auto_add_user of this UpdateCloud.  # noqa: E501
        :rtype: bool
        """
        return self._auto_add_user

    @auto_add_user.setter
    def auto_add_user(self, auto_add_user):
        """Sets the auto_add_user of this UpdateCloud.

        Whether all users in the organization should be automatically added to this cloud. This field is only relevant for organizations with cloud isolation enabled, because all users in the organization automatically have access to all clouds if cloud isolation is not enabled.  # noqa: E501

        :param auto_add_user: The auto_add_user of this UpdateCloud.  # noqa: E501
        :type: bool
        """

        self._auto_add_user = auto_add_user

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UpdateCloud):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UpdateCloud):
            return True

        return self.to_dict() != other.to_dict()



from unittest.mock import Mock, patch

from anyscale.formatters.service_formatter import format_service_config
from anyscale.models.service_model import ServiceConfig


def test_format_service_v2_config():
    """
    Tests that for a Service config with ray_serve_config field set,
    the formatter method will return both Service v1 and v2 configs.
    """
    ray_serve_config = {"runtime_env": {"pip": ["requests"], "working_dir": "."}}
    project_id = "test_project_id"
    build_id = "test_build_id"
    compute_config_id = "test_compute_config_id"
    version = "test_version"
    canary_percent = 60
    config_dict = {
        "name": "test_service",
        "ray_serve_config": ray_serve_config,
        "project_id": project_id,
        "build_id": build_id,
        "compute_config_id": compute_config_id,
        "version": version,
        "canary_percent": canary_percent,
    }
    mock_validate_successful_build = Mock()
    with patch.multiple(
        "anyscale.models.job_model",
        validate_successful_build=mock_validate_successful_build,
    ):
        service_config = ServiceConfig.parse_obj(config_dict)

    apply_service_config = format_service_config(service_config)

    assert apply_service_config.project_id == project_id
    assert apply_service_config.ray_serve_config == ray_serve_config
    assert apply_service_config.version == version
    assert apply_service_config.canary_percent == canary_percent



import asyncio
from typing import List, Optional, Tuple

from anyscale.api_utils.exceptions.log_retrieval_errors import (
    wrap_as_unsupported_log_retrieval_method_error,
    wrap_job_run_log_not_retrievable_on_active_cluster_error,
)
from anyscale.api_utils.job_util import _get_job_run_id
from anyscale.api_utils.logs_util import (
    _download_log_from_ray_json_response,
    _download_logs_concurrently,
    _remove_ansi_escape_sequences,
)
from anyscale.cli_logger import LogsLogger
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.base_job_status import BaseJobStatus
from anyscale.controllers.logs_controller import DEFAULT_PARALLELISM
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi as BaseApi
from anyscale.sdk.anyscale_client.models.log_download_result import LogDownloadResult


async def _get_logs_from_active_job_run(
    base_api: BaseApi,
    job_run_id: str,
    raise_connection_issue_as_cli_error: bool = False,
    remove_escape_chars: bool = True,
) -> str:
    """
    Retrieves logs directly from the cluster (assumed to be active) that the job run is on.

    Raises:
        ExpectedLogRetrievalError: Reasons include:
            a. If the cluster is running < Ray 2.0
            b. The Anyscale SDK/CLI client cannot directly connect to the cluster
                (due to VPN issues or high load on the cluster)
    """
    # 1. Get the Ray API URL where it can stream (ie. download from active Ray cluster) job logs.
    # Raises ExpectedLogRetrievalError on a.
    with wrap_as_unsupported_log_retrieval_method_error():
        job_logs_url: str = base_api.get_job_logs_stream(
            job_id=job_run_id
        ).result.http_url
    # 2. Call the Ray API URL and parse the logs from the JSON response
    # Raises ExpectedLogRetrievalError on b.
    with wrap_job_run_log_not_retrievable_on_active_cluster_error(
        job_run_id, raise_connection_issue_as_cli_error
    ):
        logs = await _download_log_from_ray_json_response(job_logs_url)
        if remove_escape_chars:
            logs = _remove_ansi_escape_sequences(logs)
    return logs


async def _get_logs_from_running_production_job(
    base_api: BaseApi, production_job_id: str, remove_escape_chars: bool = True,
) -> str:
    """
    Retrieves logs directly from the cluster (assumed to be active) that the **job** is on.

    This is done by determining the last job run of the job and calling
    `_get_logs_from_active_job_run()` above.

    Raises:
        NoJobRunError: If a job run hasn't been created yet.
        ExpectedLogRetrievalError: See `_get_logs_from_active_job_run`
    """
    last_job_run_id = _get_job_run_id(base_api, job_id=production_job_id)
    return await _get_logs_from_active_job_run(
        base_api, last_job_run_id, remove_escape_chars
    )


async def _get_job_logs_from_storage_bucket(
    base_api: BaseApi,
    *,
    job_id: Optional[str] = None,
    job_run_id: Optional[str] = None,
    parallelism: int = DEFAULT_PARALLELISM,
    remove_escape_chars: bool = True,
) -> str:
    """Retrieves logs directly from the storage bucket.
    Currently, this is supported only for AWS clouds (through S3)

    Args:
        parallelism (int, optional): Defaults to 10 (AWS S3's default `max_concurrent_requests` value)
            There is no documented upper limit. However, please set a reasonable value.
        remove_escape_chars (bool, optional): Removes ANSI escape sequences from the logs, which are
            commonly used to color terminal output. Defaults to True.

    Raises:
        NoJobRunError: If a job run hasn't been created yet.
        ExpectedLogRetrievalError: If the job is on a cloud that does not support log download
            (such as GCP), or using the existing Anyscale V1 implementation.
    """
    job_run_id = _get_job_run_id(base_api, job_id=job_id, job_run_id=job_run_id)

    with wrap_as_unsupported_log_retrieval_method_error():
        log_download_result: LogDownloadResult = base_api.get_job_logs_download(
            job_id=job_run_id, all_logs=True
        ).result

    all_log_chunk_urls: List[str] = [chunk.chunk_url for chunk in log_download_result.log_chunks]  # type: ignore
    logs = await _download_logs_concurrently(all_log_chunk_urls, parallelism)
    if remove_escape_chars:
        logs = _remove_ansi_escape_sequences(logs)
    return logs


_JOB_RUN_TERMINAL_STATES = {
    BaseJobStatus.SUCCEEDED,
    BaseJobStatus.STOPPED,
    BaseJobStatus.COMPLETED,
    BaseJobStatus.FAILED,
    BaseJobStatus.UNKNOWN,
}


async def _get_job_logs_from_storage_bucket_streaming(
    api: DefaultApi,
    log: LogsLogger,
    *,
    next_page_token: Optional[str] = None,
    job_run_id: Optional[str] = None,
    parallelism: int = DEFAULT_PARALLELISM,
    remove_escape_chars: bool = True,
    cluster_journal_events_start_line: int = 0,
) -> Tuple[str, Optional[str], int, bool]:
    """Retrieves logs directly from the storage bucket.
    Also fetches cluster journal logs and print them to stderr inline with job logs.

    Args:
        parallelism (int, optional): Defaults to 10 (AWS S3's default `max_concurrent_requests` value)
            There is no documented upper limit. However, please set a reasonable value.
        remove_escape_chars (bool, optional): Removes ANSI escape sequences from the logs, which are
            commonly used to color terminal output. Defaults to True.

    Returns (logs, next_page_token, cluster_journal_events_new_start_line, is_job_run_terminated):
        Returns the logs output and the next page token for fetching additional logs.
        cluster_journal_events_new_start_line is the new start line for fetching cluster journal events.
        is_job_run_terminated is True if the job run is terminated, False otherwise.

    Raises:
        NoJobRunError: If a job run hasn't been created yet.
        ExpectedLogRetrievalError: If the job is on a cloud that does not support log download
            (such as GCP), or using the existing Anyscale V1 implementation.
    """
    # First fetch both the job info and the log chunks in parallel

    job_run_thread = api.get_decorated_job_api_v2_decorated_jobs_job_id_get(
        job_id=job_run_id, async_req=True
    )

    log_download_result_thread = api.get_job_logs_download_v2_api_v2_logs_job_logs_download_v2_job_id_get(
        job_id=job_run_id, next_page_token=next_page_token, async_req=True
    )

    # Once job info is fetched, fetch the cluster journal events
    job_run = job_run_thread.get().result
    cluster_id = job_run.cluster.id

    cluster_journal_events_thread = api.get_startup_logs_api_v2_sessions_session_id_startup_logs_get(
        cluster_id,
        start_line=cluster_journal_events_start_line,
        end_line=10000000,
        async_req=True,
    )

    # Wait for the log chunks API and start downloading the chunks in parallel
    log_download_result: LogDownloadResult = log_download_result_thread.get().result
    all_log_chunk_urls: List[str] = [chunk.chunk_url for chunk in log_download_result.log_chunks]  # type: ignore
    logs_task = asyncio.create_task(
        _download_logs_concurrently(
            all_log_chunk_urls,
            parallelism,
            bearer_token=log_download_result.bearer_token,
        )
    )

    # Print out the cluster journal events
    cluster_journal_events = cluster_journal_events_thread.get().result
    lines = cluster_journal_events.lines.splitlines()
    new_cluster_journal_events_start_line = (
        cluster_journal_events.start_line + cluster_journal_events.num_lines
    )
    if cluster_journal_events_start_line == 0:
        # When fetching the inital cluster journal events, we will just show the latest 10 events
        # to avoid flooding the output with too many events. In particular, after a job is
        # already running.
        lines = lines[-10:]
    for line in lines:
        log.info(line)

    # Finally wait for all the log chunks to download and return all the logs
    logs = await logs_task
    if remove_escape_chars:
        logs = _remove_ansi_escape_sequences(logs)
    new_next_page_token = (
        log_download_result.next_page_token
        if len(log_download_result.log_chunks) > 0
        else next_page_token
    )
    return (
        logs,
        new_next_page_token,
        new_cluster_journal_events_start_line,
        job_run.status in _JOB_RUN_TERMINAL_STATES,
    )



from anyscale.shared_anyscale_utils.utils.ray_semver import ray_semver_compare


def get_correct_name_for_base_image_name(base_image_name: str) -> str:
    """
    Modify base_image_name to be a correct base image name if it is not.

    This function checks the base_image_name, if it looks correct, returns it.
    If not, try to modify it to be a correct base image name. This can be used
    as a hint for users on what they might want to use when the given name
    is not a right one.

    Args:
        base_image_name (str): The name of the base image to be checked.

    Returns:
        str: The modified base image name based on version and naming conventions.

    Example:
        To check a base image name, you can call this function like this:
        >>> get_correct_name_for_base_image_name("anyscale/ray:2.7.0-py38-cuda121"), it should be modified to "anyscale/ray:2.7.0optimized-py38-cuda121"
    """
    # to extract "2.7.0", "2.7.0oss", "2.7.0optimized", "2.0.1rc" as ray_version
    ray_version = get_ray_version(base_image_name)

    # If it has an explicit suffix, respect it.
    if ray_version.endswith(("optimized", "oss")) or "rc" in ray_version:
        return base_image_name

    # Ray 2.7.x images do not exist on anyscale. Only "optimized" version exists.
    if (
        ray_semver_compare(ray_version, "2.7.0") >= 0
        and ray_semver_compare(ray_version, "2.8.0") < 0
    ):  # ray_version ~= 2.7.?
        return base_image_name.replace(ray_version, f"{ray_version}optimized")

    return base_image_name


def get_ray_version(base_image: str) -> str:
    """Returns the Ray version in use based on the base image.

    Args:
        base_image: e.g. anyscale/ray-ml:1.9.0-cpu

    Returns:
        The ray version, e.g. 1.9.0.
    """
    # e.g. 1.9.0-cpu
    image_version = base_image.split(":")[-1]
    # e.g. 1.9.0
    ray_version = image_version.split("-")[0]
    return ray_version



import itertools
import os
import re

from setuptools import Distribution, find_packages, setup


# mypy: ignore-errors


def find_version(path):
    with open(path) as f:
        match = re.search(
            r"^__version__ = ['\"]([^'\"]*)['\"]", f.read(), re.MULTILINE,
        )
        if match:
            return match.group(1)
        raise RuntimeError("Unable to find version string.")


def read_requirements(path):
    with open(path) as f:
        return [line.strip() for line in f.readlines()]


class BinaryDistribution(Distribution):
    def is_pure(self):
        return True

    def has_ext_modules(self):
        return True


other_extras_requires = {
    "gcp": [
        # Python 3.11 does not work well with protobuf<4 ,
        # when google-cloud-* and google-api packages are required.
        "protobuf < 4 ; python_version < '3.11'",
        "google-api-python-client",
        "google-cloud-secret-manager",
        "google-cloud-compute",
        "google-cloud-resource-manager",
        "google-cloud-filestore",
        "google-cloud-storage",
        "google-cloud-redis",
    ]
}


# Flattens the values (List[str]) of each extra_requires into one List[str]
# Currently this assumes all packages defined in each extra is mutually exclusive
all_extras_requires_dependencies = ["ray>=1.4.0"] + list(
    itertools.chain.from_iterable(other_extras_requires.values())
)
all_extras_requires = {"all": all_extras_requires_dependencies}

internal_extras_requires_dependencies = [
    "terminado==0.10.1",
    "tornado",
] + all_extras_requires_dependencies
internal_extras_requires = {"backend": internal_extras_requires_dependencies}

# If adding new webterminal deps,
# Update backend/server/services/application_templates_service.py
# to prevent users from uninstalling them.
extras_require = {
    **internal_extras_requires,
    **other_extras_requires,
    **all_extras_requires,
}


def package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


VERSION_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "anyscale", "version.py"
)
REQUIREMENTS_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "requirements.in"
)

with open("README.md") as fh:
    long_description = fh.read()


setup(
    name="anyscale",
    version=find_version(VERSION_PATH),
    author="Anyscale Inc.",
    description=("Command Line Interface for Anyscale"),
    long_description=long_description,
    packages=[*find_packages(exclude="tests")],
    distclass=BinaryDistribution,
    setup_requires=["setuptools_scm"],
    package_data={
        "": [
            "anyscale/webterminal/bash-preexec.sh",
            "requirements.in",
            "*.yaml",
            "*.json",
            *package_files("anyscale/client"),
            *package_files("anyscale/sdk"),
        ],
    },
    install_requires=read_requirements(REQUIREMENTS_PATH),
    extras_require=extras_require,
    entry_points={"console_scripts": ["anyscale=anyscale.scripts:main"]},
    include_package_data=True,
    zip_safe=False,
    license_files=["LICENSE", "NOTICE"],
    license="AS License",
)



from functools import partial
import json
import os
from typing import Any, MutableMapping

from anyscale.anyscale_pydantic import BaseModel
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.util import get_working_dir


class SSHKeyInfo(BaseModel):
    user: str
    key_path: str


class WorkspaceCommandContext(BaseModel):
    ssh: SSHKeyInfo
    working_dir: str
    head_ip: str
    project_id: str


def _get_ssh_key_info(
    cluster_config: MutableMapping[str, Any],
    session_id: str,
    api_client: DefaultApi,
    ssh_dir: str = "~/.ssh",
) -> SSHKeyInfo:
    # TODO (yiran): cleanup SSH keys if session no longer exists.
    def _write_ssh_key(name: str, ssh_key: str) -> str:
        key_path = os.path.join(os.path.expanduser(ssh_dir), f"{name}.pem")
        os.makedirs(os.path.dirname(key_path), exist_ok=True)

        with open(key_path, "w", opener=partial(os.open, mode=0o600)) as f:
            f.write(ssh_key)

        return key_path

    ssh_key = api_client.get_session_ssh_key_api_v2_sessions_session_id_ssh_key_get(
        session_id
    ).result

    key_path = _write_ssh_key(ssh_key.key_name, ssh_key.private_key)

    return SSHKeyInfo(user=cluster_config["auth"]["ssh_user"], key_path=key_path)


def extract_workspace_parameters_from_cluster_config(
    cluster_id: str, project_id: str, api_client: DefaultApi, ssh_dir: str,
) -> WorkspaceCommandContext:
    """
    This function pulls relevant fields from the legacy cluster config
    to populate a `WorkspaceCommandContext`.

    TODO: Remove any reliance on cluster configs.
    NOTE: Any additional fields that need to go into the WorkspaceCommandContext
    should be passed via actual API calls, and not pulled from cluster configs.
    """

    cluster_config = json.loads(
        api_client.get_session_cluster_config_api_v2_sessions_session_id_cluster_config_get(
            cluster_id
        ).result.config_with_defaults
    )

    head_ip = api_client.get_session_head_ip_api_v2_sessions_session_id_head_ip_get(
        cluster_id
    ).result.head_ip

    ssh_info = _get_ssh_key_info(cluster_config, cluster_id, api_client, ssh_dir)
    return WorkspaceCommandContext(
        ssh=ssh_info,
        project_id=project_id,
        head_ip=head_ip,
        working_dir=get_working_dir(cluster_config, project_id, api_client) + "/",
    )
